@@ -2,34 +2,32 @@
 A collection of most basic EOTasks
 
 Credits:
-Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
-Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
-Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)
+Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
+Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Tomislav Slijepčević, Nejc Vesel, Jovan Višnjić (Sinergise)
+Copyright (c) 2017-2022 Anže Zupanc (Sinergise)
+Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
+Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)
 
 This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
 import copy
-import warnings
-from abc import abstractmethod
+from abc import abstractmethod, ABCMeta
 
 import fs
 import numpy as np
 
 from .eodata import EOPatch
 from .eotask import EOTask
-from .fs_utils import get_filesystem
-from .utilities import renamed_and_deprecated
-
-warnings.simplefilter('default', DeprecationWarning)
+from .utils.fs import get_filesystem
 
 
 class CopyTask(EOTask):
     """Makes a shallow copy of the given EOPatch.
 
     It copies feature type dictionaries but not the data itself.
-
     """
+
     def __init__(self, features=...):
         """
         :param features: A collection of features or feature types that will be copied into a new EOPatch.
@@ -42,15 +40,15 @@ class CopyTask(EOTask):
 
 
 class DeepCopyTask(CopyTask):
-    """ Makes a deep copy of the given EOPatch.
-    """
+    """Makes a deep copy of the given EOPatch."""
+
     def execute(self, eopatch):
         return eopatch.__deepcopy__(features=self.features)
 
 
-class IOTask(EOTask):
-    """ An abstract Input/Output task that can handle a path and a filesystem object
-    """
+class IOTask(EOTask, metaclass=ABCMeta):
+    """An abstract Input/Output task that can handle a path and a filesystem object."""
+
     def __init__(self, path, filesystem=None, create=False, config=None):
         """
         :param path: root path where all EOPatches are saved
@@ -69,12 +67,11 @@ class IOTask(EOTask):
         self._create = create
         self.config = config
 
-        self.filesystem_path = '/' if self._filesystem is None else self.path
+        self.filesystem_path = "/" if self._filesystem is None else self.path
 
     @property
     def filesystem(self):
-        """ A filesystem property that either initializes a new object or returns an existing one
-        """
+        """A filesystem property that either initializes a new object or returns an existing one."""
         if self._filesystem is None:
             return get_filesystem(self.path, create=self._create, config=self.config)
 
@@ -82,14 +79,12 @@ class IOTask(EOTask):
 
     @abstractmethod
     def execute(self, *eopatches, **kwargs):
-        """ Implement execute function
-        """
-        raise NotImplementedError
+        """Implement execute function"""
 
 
 class SaveTask(IOTask):
-    """ Saves the given EOPatch to a filesystem
-    """
+    """Saves the given EOPatch to a filesystem."""
+
     def __init__(self, path, filesystem=None, config=None, **kwargs):
         """
         :param path: root path where all EOPatches are saved
@@ -112,7 +107,7 @@ class SaveTask(IOTask):
         self.kwargs = kwargs
         super().__init__(path, filesystem=filesystem, create=True, config=config)
 
-    def execute(self, eopatch, *, eopatch_folder=''):
+    def execute(self, eopatch, *, eopatch_folder=""):
         """Saves the EOPatch to disk: `folder/eopatch_folder`.
 
         :param eopatch: EOPatch which will be saved
@@ -129,8 +124,8 @@ class SaveTask(IOTask):
 
 
 class LoadTask(IOTask):
-    """ Loads an EOPatch from a filesystem
-    """
+    """Loads an EOPatch from a filesystem."""
+
     def __init__(self, path, filesystem=None, config=None, **kwargs):
         """
         :param path: root directory where all EOPatches are saved
@@ -149,7 +144,7 @@ class LoadTask(IOTask):
         self.kwargs = kwargs
         super().__init__(path, filesystem=filesystem, create=False, config=config)
 
-    def execute(self, *, eopatch_folder=''):
+    def execute(self, *, eopatch_folder=""):
         """Loads the EOPatch from disk: `folder/eopatch_folder`.
 
         :param eopatch_folder: name of EOPatch folder containing data
@@ -163,14 +158,14 @@ class LoadTask(IOTask):
 
 
 class AddFeatureTask(EOTask):
-    """Adds a feature to the given EOPatch.
-    """
+    """Adds a feature to the given EOPatch."""
+
     def __init__(self, feature):
         """
         :param feature: Feature to be added
         :type feature: (FeatureType, feature_name) or FeatureType
         """
-        self.feature_type, self.feature_name = next(self._parse_features(feature)())
+        self.feature_type, self.feature_name = self.parse_feature(feature)
 
     def execute(self, eopatch, data):
         """Returns the EOPatch with added features.
@@ -191,14 +186,14 @@ class AddFeatureTask(EOTask):
 
 
 class RemoveFeatureTask(EOTask):
-    """Removes one or multiple features from the given EOPatch.
-    """
+    """Removes one or multiple features from the given EOPatch."""
+
     def __init__(self, features):
         """
         :param features: A collection of features to be removed.
         :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
-        self.feature_gen = self._parse_features(features)
+        self.feature_parser = self.get_feature_parser(features)
 
     def execute(self, eopatch):
         """Returns the EOPatch with removed features.
@@ -208,7 +203,7 @@ class RemoveFeatureTask(EOTask):
         :return: input EOPatch without the specified feature
         :rtype: EOPatch
         """
-        for feature_type, feature_name in list(self.feature_gen(eopatch)):
+        for feature_type, feature_name in self.feature_parser.get_features(eopatch):
             if feature_name is ...:
                 eopatch.reset_feature_type(feature_type)
             else:
@@ -218,14 +213,14 @@ class RemoveFeatureTask(EOTask):
 
 
 class RenameFeatureTask(EOTask):
-    """Renames one or multiple features from the given EOPatch.
-    """
+    """Renames one or multiple features from the given EOPatch."""
+
     def __init__(self, features):
         """
         :param features: A collection of features to be renamed.
         :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
-        self.feature_gen = self._parse_features(features, new_names=True)
+        self.feature_parser = self.get_feature_parser(features)
 
     def execute(self, eopatch):
         """Returns the EOPatch with renamed features.
@@ -235,7 +230,7 @@ class RenameFeatureTask(EOTask):
         :return: input EOPatch with the renamed features
         :rtype: EOPatch
         """
-        for feature_type, feature_name, new_feature_name in self.feature_gen(eopatch):
+        for feature_type, feature_name, new_feature_name in self.feature_parser.get_renamed_features(eopatch):
             eopatch[feature_type][new_feature_name] = eopatch[feature_type][feature_name]
             del eopatch[feature_type][feature_name]
 
@@ -243,8 +238,7 @@ class RenameFeatureTask(EOTask):
 
 
 class DuplicateFeatureTask(EOTask):
-    """Duplicates one or multiple features in an EOPatch.
-    """
+    """Duplicates one or multiple features in an EOPatch."""
 
     def __init__(self, features, deep_copy=False):
         """
@@ -253,7 +247,7 @@ class DuplicateFeatureTask(EOTask):
         :param deep_copy: Make a deep copy of feature's data if set to true, else just assign it.
         :type deep_copy: bool
         """
-        self.feature_gen = self._parse_features(features, new_names=True)
+        self.feature_parser = self.get_feature_parser(features)
         self.deep = deep_copy
 
     def execute(self, eopatch):
@@ -267,7 +261,7 @@ class DuplicateFeatureTask(EOTask):
             already existing feature name.
         """
 
-        for feature_type, feature_name, new_feature_name in self.feature_gen(eopatch):
+        for feature_type, feature_name, new_feature_name in self.feature_parser.get_renamed_features(eopatch):
             if new_feature_name in eopatch[feature_type]:
                 raise ValueError(f"A feature named '{new_feature_name}' already exists.")
 
@@ -280,7 +274,7 @@ class DuplicateFeatureTask(EOTask):
 
 
 class InitializeFeatureTask(EOTask):
-    """ Initializes the values of a feature.
+    """Initializes the values of a feature.
 
     Example:
 
@@ -291,6 +285,7 @@ class InitializeFeatureTask(EOTask):
         # Initialize data of the same shape as (FeatureType.DATA, 'data1')
         InitializeFeature((FeatureType.MASK, 'mask1'), shape=(FeatureType.DATA, 'data1'), init_value=1)
     """
+
     def __init__(self, features, shape, init_value=0, dtype=np.uint8):
         """
         :param features: A collection of features to initialize.
@@ -304,10 +299,10 @@ class InitializeFeatureTask(EOTask):
         :raises ValueError: Raises an exception when passing the wrong shape argument.
         """
 
-        self.features = self._parse_features(features)
+        self.features = self.parse_features(features)
 
         try:
-            self.shape_feature = next(self._parse_features(shape)())
+            self.shape_feature = self.parse_feature(shape)
         except ValueError:
             self.shape_feature = None
 
@@ -339,8 +334,8 @@ class InitializeFeatureTask(EOTask):
 
 
 class MoveFeatureTask(EOTask):
-    """ Task to copy/deepcopy fields from one eopatch to another.
-    """
+    """Task to copy/deepcopy fields from one EOPatch to another."""
+
     def __init__(self, features, deep_copy=False):
         """
         :param features: A collection of features to be moved.
@@ -348,7 +343,7 @@ class MoveFeatureTask(EOTask):
         :param deep_copy: Make a deep copy of feature's data if set to true, else just assign it.
         :type deep_copy: bool
         """
-        self.feature_gen = self._parse_features(features)
+        self.feature_parser = self.get_feature_parser(features)
         self.deep = deep_copy
 
     def execute(self, src_eopatch, dst_eopatch):
@@ -361,7 +356,7 @@ class MoveFeatureTask(EOTask):
         :rtype: EOPatch
         """
 
-        for feature in self.feature_gen(src_eopatch):
+        for feature in self.feature_parser.get_features(src_eopatch):
             if self.deep:
                 dst_eopatch[feature] = copy.deepcopy(src_eopatch[feature])
             else:
@@ -371,43 +366,44 @@ class MoveFeatureTask(EOTask):
 
 
 class MapFeatureTask(EOTask):
-    """ Applies a function to each feature in input_features of a patch and stores the results in a set of
-        output_features.
+    """Applies a function to each feature in input_features of a patch and stores the results in a set of
+    output_features.
 
-        Example using inheritance:
+    Example using inheritance:
 
-        .. code-block:: python
+    .. code-block:: python
 
-            class MultiplyFeatures(MapFeatureTask):
-                def map_method(self, f):
-                    return f * 2
+        class MultiplyFeatures(MapFeatureTask):
+            def map_method(self, f):
+                return f * 2
 
-            multiply = MultiplyFeatures({FeatureType.DATA: ['f1', 'f2', 'f3']},  # input features
-                                        {FeatureType.MASK: ['m1', 'm2', 'm3']})  # output features
+        multiply = MultiplyFeatures({FeatureType.DATA: ['f1', 'f2', 'f3']},  # input features
+                                    {FeatureType.MASK: ['m1', 'm2', 'm3']})  # output features
 
-            result = multiply(patch)
+        result = multiply(patch)
 
-        Example using lambda:
+    Example using lambda:
 
-        .. code-block:: python
+    .. code-block:: python
 
-            multiply = MapFeatureTask({FeatureType.DATA: ['f1', 'f2', 'f3']},  # input features
-                                      {FeatureType.MASK: ['m1', 'm2', 'm3']},  # output features
-                                      lambda f: f*2)                           # a function to apply to each feature
+        multiply = MapFeatureTask({FeatureType.DATA: ['f1', 'f2', 'f3']},  # input features
+                                  {FeatureType.MASK: ['m1', 'm2', 'm3']},  # output features
+                                  lambda f: f*2)                           # a function to apply to each feature
 
-            result = multiply(patch)
+        result = multiply(patch)
 
-        Example using a np.max and it's kwargs passed as arguments to the MapFeatureTask:
+    Example using a np.max and it's kwargs passed as arguments to the MapFeatureTask:
 
-        .. code-block:: python
+    .. code-block:: python
 
-            maximum = MapFeatureTask((FeatureType.DATA: 'f1'),  # input features
-                                     (FeatureType.MASK, 'm1'),  # output feature
-                                     np.max,                    # a function to apply to each feature
-                                     axis=0)                    # function's kwargs
+        maximum = MapFeatureTask((FeatureType.DATA: 'f1'),  # input features
+                                 (FeatureType.MASK, 'm1'),  # output feature
+                                 np.max,                    # a function to apply to each feature
+                                 axis=0)                    # function's kwargs
 
-            result = maximum(patch)
+        result = maximum(patch)
     """
+
     def __init__(self, input_features, output_features, map_function=None, **kwargs):
         """
         :param input_features: A collection of the input features to be mapped.
@@ -418,12 +414,12 @@ class MapFeatureTask(EOTask):
         :raises ValueError: Raises an exception when passing feature collections with different lengths.
         :param kwargs: kwargs to be passed to the map function.
         """
-        self.input_features = list(self._parse_features(input_features))
-        self.output_feature = list(self._parse_features(output_features))
+        self.input_features = self.parse_features(input_features)
+        self.output_feature = self.parse_features(output_features)
         self.kwargs = kwargs
 
         if len(self.input_features) != len(self.output_feature):
-            raise ValueError('The number of input and output features must match.')
+            raise ValueError("The number of input and output features must match.")
 
         self.function = map_function if map_function else self.map_method
 
@@ -443,11 +439,11 @@ class MapFeatureTask(EOTask):
         """
         A function that will be applied to the input features.
         """
-        raise NotImplementedError('map_method should be overridden.')
+        raise NotImplementedError("map_method should be overridden.")
 
 
 class ZipFeatureTask(EOTask):
-    """ Passes a set of input_features to a function, which returns a single features as a result and stores it in
+    """Passes a set of input_features to a function, which returns a single features as a result and stores it in
     the given EOPatch.
 
         Example using inheritance:
@@ -484,6 +480,7 @@ class ZipFeatureTask(EOTask):
 
             result = maximum(patch)
     """
+
     def __init__(self, input_features, output_feature, zip_function=None, **kwargs):
         """
         :param input_features: A collection of the input features to be mapped.
@@ -493,8 +490,8 @@ class ZipFeatureTask(EOTask):
         :param zip_function: A function or lambda to be applied to the input data.
         :param kwargs: kwargs to be passed to the zip function.
         """
-        self.input_features = list(self._parse_features(input_features))
-        self.output_feature = next(self._parse_features(output_feature)())
+        self.input_features = self.parse_features(input_features)
+        self.output_feature = self.parse_feature(output_feature)
         self.function = zip_function if zip_function else self.zip_method
         self.kwargs = kwargs
 
@@ -512,23 +509,21 @@ class ZipFeatureTask(EOTask):
         return eopatch
 
     def zip_method(self, *f):
-        """ A function that will be applied to the input features if overridden.
-        """
-        raise NotImplementedError('zip_method should be overridden.')
+        """A function that will be applied to the input features if overridden."""
+        raise NotImplementedError("zip_method should be overridden.")
 
 
 class MergeFeatureTask(ZipFeatureTask):
-    """ Merges multiple features together by concatenating their data along the last axis.
-    """
+    """Merges multiple features together by concatenating their data along the last axis."""
+
     def zip_method(self, *f):
-        """Concatenates the data of features along the last axis.
-        """
+        """Concatenates the data of features along the last axis."""
         return np.concatenate(f, axis=-1)
 
 
 class ExtractBandsTask(MapFeatureTask):
-    """ Moves a subset of bands from one feature to a new one.
-    """
+    """Moves a subset of bands from one feature to a new one."""
+
     def __init__(self, input_feature, output_feature, bands):
         """
         :param input_feature: A source feature from which to take the subset of bands.
@@ -549,8 +544,8 @@ class ExtractBandsTask(MapFeatureTask):
 
 
 class CreateEOPatchTask(EOTask):
-    """Creates an EOPatch
-    """
+    """Creates an EOPatch."""
+
     def execute(self, **kwargs):
         """Returns a newly created EOPatch with the given kwargs.
 
@@ -562,10 +557,11 @@ class CreateEOPatchTask(EOTask):
 
 
 class MergeEOPatchesTask(EOTask):
-    """ Merge content from multiple EOPatches into a single EOPatch
+    """Merge content from multiple EOPatches into a single EOPatch.
 
     Check :func:`EOPatch.merge<eolearn.core.eodata.EOPatch.merge>` for more information about the merging process.
     """
+
     def __init__(self, **merge_kwargs):
         """
         :param merge_kwargs: Keyword arguments defined for `EOPatch.merge` method.
@@ -580,54 +576,6 @@ class MergeEOPatchesTask(EOTask):
         :rtype: EOPatch
         """
         if not eopatches:
-            raise ValueError('At least one EOPatch should be given')
+            raise ValueError("At least one EOPatch should be given")
 
         return eopatches[0].merge(*eopatches[1:], **self.merge_kwargs)
-
-
-@renamed_and_deprecated
-class SaveToDisk(SaveTask):
-    """ A deprecated version of SaveTask
-    """
-
-
-@renamed_and_deprecated
-class LoadFromDisk(LoadTask):
-    """ A deprecated version of LoadTask
-    """
-
-
-@renamed_and_deprecated
-class AddFeature(AddFeatureTask):
-    """ A deprecated version of AddFeatureTask
-    """
-
-
-@renamed_and_deprecated
-class RemoveFeature(RemoveFeatureTask):
-    """ A deprecated version of RemoveFeatureTask
-    """
-
-
-@renamed_and_deprecated
-class RenameFeature(RenameFeatureTask):
-    """ A deprecated version of RenameFeatureTask
-    """
-
-
-@renamed_and_deprecated
-class DuplicateFeature(DuplicateFeatureTask):
-    """ A deprecated version of DuplicateFeatureTask
-    """
-
-
-@renamed_and_deprecated
-class InitializeFeature(InitializeFeatureTask):
-    """ A deprecated version of InitializeFeatureTask
-    """
-
-
-@renamed_and_deprecated
-class MoveFeature(MoveFeatureTask):
-    """ A deprecated version of MoveFeatureTask
-    """
