@@ -2,17 +2,21 @@
 The eodata module provides core objects for handling remote sensing multi-temporal data (such as satellite imagery).
 
 Credits:
-Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
-Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
-Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)
+Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
+Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Tomislav Slijepčević, Nejc Vesel, Jovan Višnjić (Sinergise)
+Copyright (c) 2017-2022 Anže Zupanc (Sinergise)
+Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
+Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)
 
 This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
+from __future__ import annotations
+
 import logging
-import warnings
 import copy
 import datetime
+from typing import Tuple, Union, List, Optional, TYPE_CHECKING
 
 import attr
 import dateutil.parser
@@ -24,15 +28,22 @@ from sentinelhub import BBox, CRS
 from .constants import FeatureType, OverwritePermission
 from .eodata_io import save_eopatch, load_eopatch, FeatureIO
 from .eodata_merge import merge_eopatches
-from .fs_utils import get_filesystem
-from .utilities import deep_eq, FeatureParser
+from .utils.fs import get_filesystem
+from .utils.common import deep_eq, is_discrete_type
+from .utils.parsing import parse_features
 
 
 LOGGER = logging.getLogger(__name__)
-warnings.simplefilter('default', DeprecationWarning)
 
 MAX_DATA_REPR_LEN = 100
 
+if TYPE_CHECKING:
+    try:
+        from eolearn.visualization import PlotBackend
+        from eolearn.visualization.eopatch_base import BasePlotConfig
+    except ImportError:
+        pass
+
 
 @attr.s(repr=False, eq=False, kw_only=True)
 class EOPatch:
@@ -54,6 +65,7 @@ class EOPatch:
     Currently the EOPatch object doesn't enforce that the length of timestamp be equal to n_times dimensions of numpy
     arrays in other attributes.
     """
+
     data = attr.ib(factory=dict)
     mask = attr.ib(factory=dict)
     scalar = attr.ib(factory=dict)
@@ -71,7 +83,7 @@ class EOPatch:
     def __setattr__(self, key, value, feature_name=None):
         """Raises TypeError if feature type attributes are not of correct type.
 
-        In case they are a dictionary they are cast to _FeatureDict class
+        In case they are a dictionary they are cast to _FeatureDict class.
         """
         if feature_name not in (None, Ellipsis) and FeatureType.has_value(key):
             self[key][feature_name] = value
@@ -85,7 +97,7 @@ class EOPatch:
 
     @staticmethod
     def _parse_feature_type_value(feature_type, value):
-        """ Checks or parses value which will be assigned to a feature type attribute of `EOPatch`. If the value
+        """Checks or parses value which will be assigned to a feature type attribute of `EOPatch`. If the value
         cannot be parsed correctly it raises an error.
 
         :raises: TypeError, ValueError
@@ -101,15 +113,18 @@ class EOPatch:
 
         if feature_type is FeatureType.TIMESTAMP:
             if isinstance(value, (tuple, list)):
-                return [timestamp if isinstance(timestamp, datetime.date) else dateutil.parser.parse(timestamp)
-                        for timestamp in value]
+                return [
+                    timestamp if isinstance(timestamp, datetime.date) else dateutil.parser.parse(timestamp)
+                    for timestamp in value
+                ]
 
-        raise TypeError(f'Attribute {feature_type} requires value of type {feature_type.type()} - '
-                        f'failed to parse given value {value}')
+        raise TypeError(
+            f"Attribute {feature_type} requires value of type {feature_type.type()} - "
+            f"failed to parse given value {value}"
+        )
 
     def __getattribute__(self, key, load=True, feature_name=None):
-        """ Handles lazy loading and it can even provide a single feature from _FeatureDict
-        """
+        """Handles lazy loading and it can even provide a single feature from _FeatureDict."""
         value = super().__getattribute__(key)
 
         if isinstance(value, FeatureIO) and load:
@@ -123,7 +138,7 @@ class EOPatch:
         return value
 
     def __getitem__(self, feature_type):
-        """ Provides features of requested feature type. It can also accept a tuple of (feature_type, feature_name)
+        """Provides features of requested feature type. It can also accept a tuple of (feature_type, feature_name).
 
         :param feature_type: Type of EOPatch feature
         :type feature_type: FeatureType or str or (FeatureType, str)
@@ -138,7 +153,7 @@ class EOPatch:
 
     def __setitem__(self, feature_type, value):
         """Sets a new dictionary / list to the given FeatureType. As a key it can also accept a tuple of
-        (feature_type, feature_name)
+        (feature_type, feature_name).
 
         :param feature_type: Type of EOPatch feature
         :type feature_type: FeatureType or str or (FeatureType, str)
@@ -154,7 +169,7 @@ class EOPatch:
         return self.__setattr__(FeatureType(feature_type).value, value, feature_name=feature_name)
 
     def __delitem__(self, feature):
-        """ Deletes the selected feature.
+        """Deletes the selected feature.
 
         :param feature: EOPatch feature
         :type feature: (FeatureType, str)
@@ -165,10 +180,9 @@ class EOPatch:
 
     @staticmethod
     def _check_tuple_key(key):
-        """ A helper function that checks a tuple, which should hold (feature_type, feature_name)
-        """
+        """A helper function that checks a tuple, which should hold (feature_type, feature_name)."""
         if len(key) != 2:
-            raise ValueError(f'Given element should be a tuple of (feature_type, feature_name), but {key} found')
+            raise ValueError(f"Given element should be a tuple of (feature_type, feature_name), but {key} found.")
 
     def __eq__(self, other):
         """True if FeatureType attributes, bbox, and timestamps of both EOPatches are equal by value."""
@@ -180,27 +194,44 @@ class EOPatch:
                 return False
         return True
 
+    def __contains__(self, feature: Union[FeatureType, Tuple[FeatureType, str]]):
+        if isinstance(feature, FeatureType):
+            return bool(self[feature])
+        if isinstance(feature, tuple) and len(feature) == 2:
+            ftype, fname = FeatureType(feature[0]), feature[1]
+            if ftype.has_dict():
+                return fname in self[ftype]
+            return bool(self[ftype])
+        raise ValueError(
+            f"Membership checking is only implemented elements of type `{FeatureType.__name__}` and for "
+            "`(feature_type, feature_name)` tuples."
+        )
+
     def __add__(self, other):
-        """ Merges two EOPatches into a new EOPatch
-        """
+        """Merges two EOPatches into a new EOPatch."""
         return self.merge(other)
 
     def __repr__(self):
-        feature_repr_list = [f'{self.__class__.__name__}(']
+        feature_repr_list = []
         for feature_type in FeatureType:
             content = self[feature_type]
+            if not content:
+                continue
 
             if isinstance(content, dict) and content:
                 content_str = (
-                    '{\n    '
-                    + '\n    '.join([f'{label}: {self._repr_value(value)}' for label, value in sorted(content.items())])
-                    + '\n  }'
+                    "{\n    "
+                    + "\n    ".join([f"{label}: {self._repr_value(value)}" for label, value in sorted(content.items())])
+                    + "\n  }"
                 )
             else:
                 content_str = self._repr_value(content)
-            feature_repr_list.append(f'{feature_type.value}: {content_str}')
+            feature_repr_list.append(f"{feature_type.value}={content_str}")
 
-        return '\n  '.join(feature_repr_list) + '\n)'
+        feature_repr = "\n  ".join(feature_repr_list)
+        if feature_repr:
+            feature_repr = f"\n  {feature_repr}\n"
+        return f"{self.__class__.__name__}({feature_repr})"
 
     @staticmethod
     def _repr_value(value):
@@ -211,40 +242,39 @@ class EOPatch:
         :rtype: str
         """
         if isinstance(value, np.ndarray):
-            return f'{EOPatch._repr_value_class(value)}(shape={value.shape}, dtype={value.dtype})'
+            return f"{EOPatch._repr_value_class(value)}(shape={value.shape}, dtype={value.dtype})"
 
         if isinstance(value, gpd.GeoDataFrame):
             crs = CRS(value.crs).ogc_string() if value.crs else value.crs
-            return f'{EOPatch._repr_value_class(value)}(columns={list(value)}, length={len(value)}, crs={crs})'
+            return f"{EOPatch._repr_value_class(value)}(columns={list(value)}, length={len(value)}, crs={crs})"
 
         if isinstance(value, (list, tuple, dict)) and value:
             repr_str = str(value)
             if len(repr_str) <= MAX_DATA_REPR_LEN:
                 return repr_str
 
-            l_bracket, r_bracket = '[]' if isinstance(value, list) else '()'
+            l_bracket, r_bracket = ("[", "]") if isinstance(value, list) else ("(", ")")
             if isinstance(value, (list, tuple)) and len(value) > 2:
-                repr_str = f'{l_bracket}{repr(value[0])}, ..., {repr(value[-1])}{r_bracket}'
+                repr_str = f"{l_bracket}{repr(value[0])}, ..., {repr(value[-1])}{r_bracket}"
 
             if len(repr_str) > MAX_DATA_REPR_LEN and isinstance(value, (list, tuple)) and len(value) > 1:
-                repr_str = f'{l_bracket}{repr(value[0])}, ...{r_bracket}'
+                repr_str = f"{l_bracket}{repr(value[0])}, ...{r_bracket}"
 
             if len(repr_str) > MAX_DATA_REPR_LEN:
                 repr_str = str(type(value))
 
-            return f'{repr_str}, length={len(value)}'
+            return f"{repr_str}, length={len(value)}"
 
         return repr(value)
 
     @staticmethod
     def _repr_value_class(value):
-        """ A representation of a class of a given value
-        """
+        """A representation of a class of a given value"""
         cls = value.__class__
-        return '.'.join([cls.__module__.split('.')[0], cls.__name__])
+        return ".".join([cls.__module__.split(".")[0], cls.__name__])
 
     def __copy__(self, features=...):
-        """ Returns a new EOPatch with shallow copies of given features.
+        """Returns a new EOPatch with shallow copies of given features.
 
         :param features: A collection of features or feature types that will be copied into new EOPatch.
         :type features: object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
@@ -253,15 +283,15 @@ class EOPatch:
             features = ...
 
         new_eopatch = EOPatch()
-        for feature_type, feature_name in FeatureParser(features)(self):
-            if feature_name is ...:
-                new_eopatch[feature_type] = copy.copy(self[feature_type])
-            else:
+        for feature_type, feature_name in parse_features(features, eopatch=self):
+            if feature_type.has_dict():
                 new_eopatch[feature_type][feature_name] = self[feature_type].__getitem__(feature_name, load=False)
+            else:
+                new_eopatch[feature_type] = copy.copy(self[feature_type])
         return new_eopatch
 
     def __deepcopy__(self, memo=None, features=...):
-        """ Returns a new EOPatch with deep copies of given features.
+        """Returns a new EOPatch with deep copies of given features.
 
         :param memo: built-in parameter for memoization
         :type memo: dict
@@ -272,10 +302,8 @@ class EOPatch:
             features = ...
 
         new_eopatch = EOPatch()
-        for feature_type, feature_name in FeatureParser(features)(self):
-            if feature_name is ...:
-                new_eopatch[feature_type] = copy.deepcopy(self[feature_type], memo=memo)
-            else:
+        for feature_type, feature_name in parse_features(features, eopatch=self):
+            if feature_type.has_dict():
                 value = self[feature_type].__getitem__(feature_name, load=False)
 
                 if isinstance(value, FeatureIO):
@@ -286,11 +314,13 @@ class EOPatch:
                     value = copy.deepcopy(value, memo=memo)
 
                 new_eopatch[feature_type][feature_name] = value
+            else:
+                new_eopatch[feature_type] = copy.deepcopy(self[feature_type], memo=memo)
 
         return new_eopatch
 
     def copy(self, features=..., deep=False):
-        """ Get a copy of the current `EOPatch`.
+        """Get a copy of the current `EOPatch`.
 
         :param features: Features to be copied into a new `EOPatch`. By default all features will be copied.
         :type features: object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
@@ -304,67 +334,6 @@ class EOPatch:
             return self.__deepcopy__(features=features)
         return self.__copy__(features=features)
 
-    def remove_feature(self, feature_type, feature_name):
-        """Removes the feature ``feature_name`` from dictionary of ``feature_type``.
-
-        :param feature_type: Enum of the attribute we're about to modify
-        :type feature_type: FeatureType
-        :param feature_name: Name of the feature of the attribute
-        :type feature_name: str
-        """
-        LOGGER.debug("Removing feature '%s' from attribute '%s'", feature_name, feature_type.value)
-
-        self._check_if_dict(feature_type)
-        if feature_name in self[feature_type]:
-            del self[feature_type][feature_name]
-
-    def add_feature(self, feature_type, feature_name, value):
-        """Sets EOPatch[feature_type][feature_name] to the given value.
-
-        :param feature_type: Type of feature
-        :type feature_type: FeatureType
-        :param feature_name: Name of the feature
-        :type feature_name: str
-        :param value: New value of the feature
-        :type value: object
-        """
-        self._check_if_dict(feature_type)
-        self[feature_type][feature_name] = value
-
-    def rename_feature(self, feature_type, feature_name, new_feature_name):
-        """Renames the feature ``feature_name`` to ``new_feature_name`` from dictionary of ``feature_type``.
-
-        :param feature_type: Enum of the attribute we're about to rename
-        :type feature_type: FeatureType
-        :param feature_name: Name of the feature of the attribute
-        :type feature_name: str
-        :param new_feature_name: New Name of the feature of the attribute
-        :type feature_name: str
-        """
-        self._check_if_dict(feature_type)
-        if feature_name != new_feature_name:
-            if feature_name in self[feature_type]:
-                LOGGER.debug("Renaming feature '%s' from attribute '%s' to '%s'",
-                             feature_name, feature_type.value, new_feature_name)
-                self[feature_type][new_feature_name] = self[feature_type][feature_name]
-                del self[feature_type][feature_name]
-            else:
-                raise ValueError(f"Feature {feature_name} from attribute {feature_type.value} does not exist!")
-        else:
-            LOGGER.debug("Feature '%s' was not renamed because new name is identical.", feature_name)
-
-    @staticmethod
-    def _check_if_dict(feature_type):
-        """Checks if the given feature type contains a dictionary and raises an error if it doesn't.
-
-        :param feature_type: Type of feature
-        :type feature_type: FeatureType
-        :raise: TypeError
-        """
-        feature_type = FeatureType(feature_type)
-        if feature_type.type() is not dict:
-            raise TypeError(f'{feature_type} does not contain a dictionary of features')
-
     def reset_feature_type(self, feature_type):
         """Resets the values of the given feature type.
 
@@ -379,32 +348,6 @@ class EOPatch:
         else:
             self[feature_type] = []
 
-    def set_bbox(self, new_bbox):
-        """
-        :param new_bbox: new bbox
-        :type: new_bbox: BBox
-        """
-        self.bbox = new_bbox
-
-    def set_timestamp(self, new_timestamp):
-        """
-        :param new_timestamp: list of dates
-        :type new_timestamp: list(str)
-        """
-        self.timestamp = new_timestamp
-
-    def get_feature(self, feature_type, feature_name=None):
-        """Returns the array of corresponding feature.
-
-        :param feature_type: Enum of the attribute
-        :type feature_type: FeatureType
-        :param feature_name: Name of the feature
-        :type feature_name: str
-        """
-        if feature_name is None:
-            return self[feature_type]
-        return self[feature_type][feature_name]
-
     def get_features(self):
         """Returns a dictionary of all non-empty features of EOPatch.
 
@@ -432,22 +375,22 @@ class EOPatch:
         :param feature_name: Name of the feature
         :type feature_name: str
         """
-        if feature_type.is_time_dependent() or feature_type.is_spatial():
+        if feature_type.is_temporal() or feature_type.is_spatial():
             shape = self[feature_type][feature_name].shape
-            return shape[1:3] if feature_type.is_time_dependent() else shape[0:2]
+            return shape[1:3] if feature_type.is_temporal() else shape[0:2]
 
-        raise ValueError('FeatureType used to determine the width and height of raster must be'
-                         ' time dependent or spatial.')
+        raise ValueError(
+            "FeatureType used to determine the width and height of raster must be time dependent or spatial."
+        )
 
-    def get_feature_list(self):
+    def get_feature_list(self) -> List[Union[FeatureType, Tuple[FeatureType, str]]]:
         """Returns a list of all non-empty features of EOPatch.
 
         The elements are either only FeatureType or a pair of FeatureType and feature name.
 
         :return: list of features
-        :rtype: list(FeatureType or (FeatureType, str))
         """
-        feature_list = []
+        feature_list: List[Union[FeatureType, Tuple[FeatureType, str]]] = []
         for feature_type in FeatureType:
             if feature_type.has_dict():
                 for feature_name in self[feature_type]:
@@ -456,78 +399,10 @@ class EOPatch:
                 feature_list.append(feature_type)
         return feature_list
 
-    @staticmethod
-    def concatenate(eopatch1, eopatch2):
-        """Joins all data from two EOPatches and returns a new EOPatch.
-
-        If timestamps don't match it will try to join all time-dependent features with the same name.
-
-        Note: In general the data won't be deep copied. Deep copy will only happen when merging time-dependent features
-        along time
-
-        :param eopatch1: First EOPatch
-        :type eopatch1: EOPatch
-        :param eopatch2: First EOPatch
-        :type eopatch2: EOPatch
-        :return: Joined EOPatch
-        :rtype: EOPatch
-        """
-        warnings.warn('EOPatch.concatenate is deprecated, use a more general EOPatch.merge method instead',
-                      DeprecationWarning)
-
-        eopatch_content = {}
-
-        timestamps_exist = eopatch1.timestamp and eopatch2.timestamp
-        timestamps_match = timestamps_exist and deep_eq(eopatch1.timestamp, eopatch2.timestamp)
-
-        # if not timestamps_match and timestamps_exist and eopatch1.timestamp[-1] >= eopatch2.timestamp[0]:
-        #     raise ValueError('Could not merge timestamps because any timestamp of the first EOPatch must be before '
-        #                      'any timestamp of the second EOPatch')
-
-        for feature_type in FeatureType:
-            if feature_type.has_dict():
-                eopatch_content[feature_type.value] = {**eopatch1[feature_type], **eopatch2[feature_type]}
-
-                for feature_name in eopatch1[feature_type].keys() & eopatch2[feature_type].keys():
-                    data1 = eopatch1[feature_type][feature_name]
-                    data2 = eopatch2[feature_type][feature_name]
-
-                    if feature_type.is_time_dependent() and not timestamps_match:
-                        eopatch_content[feature_type.value][feature_name] = EOPatch.concatenate_data(data1, data2)
-                    elif not deep_eq(data1, data2):
-                        raise ValueError(f'Could not merge ({feature_type}, {feature_name}) feature because values '
-                                         'differ')
-
-            elif feature_type is FeatureType.TIMESTAMP and timestamps_exist and not timestamps_match:
-                eopatch_content[feature_type.value] = eopatch1[feature_type] + eopatch2[feature_type]
-            else:
-                if not eopatch1[feature_type] or deep_eq(eopatch1[feature_type], eopatch2[feature_type]):
-                    eopatch_content[feature_type.value] = copy.copy(eopatch2[feature_type])
-                elif not eopatch2[feature_type]:
-                    eopatch_content[feature_type.value] = copy.copy(eopatch1[feature_type])
-                else:
-                    raise ValueError(f'Could not merge {feature_type} feature because values differ')
-
-        return EOPatch(**eopatch_content)
-
-    @staticmethod
-    def concatenate_data(data1, data2):
-        """A method that concatenates two numpy array along first axis.
-
-        :param data1: Numpy array of shape (times1, height, width, n_features)
-        :type data1: numpy.ndarray
-        :param data2: Numpy array of shape (times2, height, width, n_features)
-        :type data1: numpy.ndarray
-        :return: Numpy array of shape (times1 + times2, height, width, n_features)
-        :rtype: numpy.ndarray
-        """
-        if data1.shape[1:] != data2.shape[1:]:
-            raise ValueError('Could not concatenate data because non-temporal dimensions do not match')
-        return np.concatenate((data1, data2), axis=0)
-
-    def save(self, path, features=..., overwrite_permission=OverwritePermission.ADD_ONLY, compress_level=0,
-             filesystem=None):
-        """ Method to save an EOPatch from memory to a storage
+    def save(
+        self, path, features=..., overwrite_permission=OverwritePermission.ADD_ONLY, compress_level=0, filesystem=None
+    ):
+        """Method to save an EOPatch from memory to a storage.
 
         :param path: A location where to save EOPatch. It can be either a local path or a remote URL path.
         :type path: str
@@ -545,14 +420,20 @@ class EOPatch:
         """
         if filesystem is None:
             filesystem = get_filesystem(path, create=True)
-            path = '/'
+            path = "/"
 
-        save_eopatch(self, filesystem, path, features=features, compress_level=compress_level,
-                     overwrite_permission=OverwritePermission(overwrite_permission))
+        save_eopatch(
+            self,
+            filesystem,
+            path,
+            features=features,
+            compress_level=compress_level,
+            overwrite_permission=OverwritePermission(overwrite_permission),
+        )
 
     @staticmethod
     def load(path, features=..., lazy_loading=False, filesystem=None):
-        """ Method to load an EOPatch from a storage into memory
+        """Method to load an EOPatch from a storage into memory.
 
         :param path: A location from where to load EOPatch. It can be either a local path or a remote URL path.
         :type path: str
@@ -568,12 +449,12 @@ class EOPatch:
         """
         if filesystem is None:
             filesystem = get_filesystem(path, create=False)
-            path = '/'
+            path = "/"
 
         return load_eopatch(EOPatch(), filesystem, path, features=features, lazy_loading=lazy_loading)
 
     def merge(self, *eopatches, features=..., time_dependent_op=None, timeless_op=None):
-        """ Merge features of given EOPatches into a new EOPatch
+        """Merge features of given EOPatches into a new EOPatch.
 
         :param eopatches: Any number of EOPatches to be merged together with the current EOPatch
         :type eopatches: EOPatch
@@ -603,8 +484,9 @@ class EOPatch:
         :return: A dictionary with EOPatch features and values
         :rtype: Dict[(FeatureType, str), object]
         """
-        eopatch_content = merge_eopatches(self, *eopatches, features=features,
-                                          time_dependent_op=time_dependent_op, timeless_op=timeless_op)
+        eopatch_content = merge_eopatches(
+            self, *eopatches, features=features, time_dependent_op=time_dependent_op, timeless_op=timeless_op
+        )
 
         merged_eopatch = EOPatch()
         for feature, value in eopatch_content.items():
@@ -612,7 +494,7 @@ class EOPatch:
 
         return merged_eopatch
 
-    def time_series(self, ref_date=None, scale_time=1):
+    def get_time_series(self, ref_date=None, scale_time=1):
         """Returns a numpy array with seconds passed between the reference date and the timestamp of each image.
 
         An array is constructed as time_series[i] = (timestamp[i] - ref_date).total_seconds().
@@ -626,13 +508,14 @@ class EOPatch:
         """
 
         if not self.timestamp:
-            return None
+            return np.zeros(0, dtype=np.int64)
 
         if ref_date is None:
             ref_date = self.timestamp[0]
 
-        return np.asarray([round((timestamp - ref_date).total_seconds() / scale_time) for timestamp in self.timestamp],
-                          dtype=np.int64)
+        return np.asarray(
+            [round((timestamp - ref_date).total_seconds() / scale_time) for timestamp in self.timestamp], dtype=np.int64
+        )
 
     def consolidate_timestamps(self, timestamps):
         """Removes all frames from the EOPatch with a date not found in the provided timestamps list.
@@ -647,8 +530,9 @@ class EOPatch:
         good_timestamp_idxs = [idx for idx, _ in enumerate(self.timestamp) if idx not in remove_from_patch_idxs]
         good_timestamps = [date for idx, date in enumerate(self.timestamp) if idx not in remove_from_patch_idxs]
 
-        for feature_type in [feature_type for feature_type in FeatureType if (feature_type.is_time_dependent() and
-                                                                              feature_type.has_dict())]:
+        for feature_type in [
+            feature_type for feature_type in FeatureType if (feature_type.is_temporal() and feature_type.has_dict())
+        ]:
 
             for feature_name, value in self[feature_type].items():
                 if isinstance(value, np.ndarray):
@@ -659,40 +543,52 @@ class EOPatch:
         self.timestamp = good_timestamps
         return remove_from_patch
 
-    def plot(self, feature, rgb=None, rgb_factor=3.5, vdims=None, timestamp_column='TIMESTAMP',
-             geometry_column='geometry', pixel=False, mask=None):
-        """ Plots eopatch features
+    def plot(
+        self,
+        feature,
+        *,
+        times: Union[List[int], slice, None] = None,
+        channels: Union[List[int], slice, None] = None,
+        channel_names: Optional[List[str]] = None,
+        rgb: Optional[Tuple[int, int, int]] = None,
+        backend: Union[str, PlotBackend] = "matplotlib",
+        config: Optional[BasePlotConfig] = None,
+        **kwargs,
+    ) -> object:
+        """Plots an `EOPatch` feature.
 
-        :param feature: feature of eopatch
-        :type feature: (FeatureType, str)
-        :param rgb: indexes of bands to create rgb image from
-        :type rgb: [int, int, int]
-        :param rgb_factor: factor for rgb bands multiplication
-        :type rgb_factor: float
-        :param vdims: value dimension for vector data
-        :type vdims: str
-        :param timestamp_column: name of the timestamp column, valid for vector data
-        :type timestamp_column: str
-        :param geometry_column: name of the geometry column, valid for vector data
-        :type geometry_column: str
-        :param pixel: plot values through time for one pixel
-        :type pixel: bool
-        :param mask: where eopatch[FeatureType.MASK] == False, value = 0
-        :type mask: str
-        :return: plot
-        :rtype: holovies/bokeh
+        :param feature: A feature in the `EOPatch`.
+        :param times: A list or a slice of indices on temporal axis to be used for plotting. If not provided all
+            indices will be used.
+        :param channels: A list or a slice of indices on channels axis to be used for plotting. If not provided all
+            indices will be used.
+        :param channel_names: Names of channels of the last dimension in the given raster feature.
+        :param rgb: If provided, it should be a list of 3 indices of RGB channels to be plotted. It will plot only RGB
+            images with these channels. This only works for raster features with spatial dimension.
+        :param backend: A type of plotting backend.
+        :param config: A configuration object with advanced plotting parameters.
+        :param kwargs: Parameters that are specific to a specified plotting backend.
+        :return: A plot object that depends on the backend used.
         """
         # pylint: disable=import-outside-toplevel,raise-missing-from
         try:
-            from eolearn.visualization import EOPatchVisualization
+            from eolearn.visualization.eopatch import plot_eopatch
         except ImportError:
-            raise RuntimeError('Subpackage eo-learn-visualization has to be installed with an option [FULL] in order '
-                               'to use plot method')
+            raise RuntimeError(
+                "Subpackage eo-learn-visualization has to be installed in order to use EOPatch visualization method"
+            )
 
-        vis = EOPatchVisualization(self, feature=feature, rgb=rgb, rgb_factor=rgb_factor, vdims=vdims,
-                                   timestamp_column=timestamp_column, geometry_column=geometry_column,
-                                   pixel=pixel, mask=mask)
-        return vis.plot()
+        return plot_eopatch(
+            self,
+            feature=feature,
+            times=times,
+            channels=channels,
+            channel_names=channel_names,
+            rgb=rgb,
+            backend=backend,
+            config=config,
+            **kwargs,
+        )
 
 
 class _FeatureDict(dict):
@@ -701,7 +597,8 @@ class _FeatureDict(dict):
     It checks that features have a correct and dimension. It also supports lazy loading by accepting a function as a
     feature value, which is then called when the feature is accessed.
     """
-    FORBIDDEN_CHARS = {'.', '/', '\\', '|', ';', ':', '\n', '\t'}
+
+    FORBIDDEN_CHARS = {".", "/", "\\", "|", ";", ":", "\n", "\t"}
 
     def __init__(self, feature_dict, feature_type):
         """
@@ -720,24 +617,25 @@ class _FeatureDict(dict):
             self[feature_name] = value
 
     def __setitem__(self, feature_name, value):
-        """ Before setting value to the dictionary it checks that value is of correct type and dimension and tries to
+        """Before setting value to the dictionary it checks that value is of correct type and dimension and tries to
         transform value in correct form.
         """
-        value = self._parse_feature_value(value)
+        value = self._parse_feature_value(value, feature_name)
         self._check_feature_name(feature_name)
         super().__setitem__(feature_name, value)
 
     def _check_feature_name(self, feature_name):
         if not isinstance(feature_name, str):
-            raise ValueError(f'Feature name must be a string but an object of type {type(feature_name)} was given.')
+            raise ValueError(f"Feature name must be a string but an object of type {type(feature_name)} was given.")
 
         for char in feature_name:
             if char in self.FORBIDDEN_CHARS:
-                raise ValueError(f"The name of feature ({self.feature_type}, {feature_name}) contains an illegal "
-                                 f"character '{char}'.")
+                raise ValueError(
+                    f"The name of feature ({self.feature_type}, {feature_name}) contains an illegal character '{char}'."
+                )
 
-        if feature_name == '':
-            raise ValueError('Feature name cannot be an empty string.')
+        if feature_name == "":
+            raise ValueError("Feature name cannot be an empty string.")
 
     def __getitem__(self, feature_name, load=True):
         """Implements lazy loading."""
@@ -750,52 +648,42 @@ class _FeatureDict(dict):
         return value
 
     def __eq__(self, other):
-        """ Compares its content against a content of another feature type dictionary
-        """
+        """Compares its content against a content of another feature type dictionary."""
         return deep_eq(self, other)
 
     def __ne__(self, other):
-        """ Compares its content against a content of another feature type dictionary
-        """
+        """Compares its content against a content of another feature type dictionary."""
         return not self.__eq__(other)
 
     def get_dict(self):
         """Returns a Python dictionary of features and value."""
         return dict(self)
 
-    def _parse_feature_value(self, value):
-        """ Checks if value fits the feature type. If not it tries to fix it or raise an error
+    def _parse_feature_value(self, value, feature_name):
+        """Checks if value fits the feature type. If not it tries to fix it or raise an error.
 
         :raises: ValueError
         """
         if isinstance(value, FeatureIO):
             return value
-        if not hasattr(self, 'ndim'):  # Because of serialization/deserialization during multiprocessing
+        if not hasattr(self, "ndim"):  # Because of serialization/deserialization during multiprocessing
             return value
 
         if self.ndim:
             if not isinstance(value, np.ndarray):
-                raise ValueError(f'{self.feature_type} feature has to be a numpy array')
+                raise ValueError(f"{self.feature_type} feature has to be a numpy array.")
             if value.ndim != self.ndim:
-                raise ValueError(f'Numpy array of {self.feature_type} feature has to have {self.ndim} '
-                                 f'dimension{"s" if self.ndim > 1 else ""}')
+                raise ValueError(
+                    f"Numpy array of {self.feature_type} feature has to have {self.ndim} "
+                    f"dimension{'s' if self.ndim > 1 else ''} but feature {feature_name} has {value.ndim}."
+                )
 
-            if self.feature_type.is_discrete():
-                if not issubclass(value.dtype.type, (np.integer, bool, np.bool_, np.bool8)):
-                    msg = (
-                        f'{self.feature_type} is a discrete feature type therefore dtype of data should be a subtype '
-                        f'of numpy.integer or numpy.bool, found type {value.dtype.type}. In the future an error will '
-                        'be raised because of this'
-                    )
-                    warnings.warn(msg, DeprecationWarning, stacklevel=3)
+            if self.feature_type.is_discrete() and not is_discrete_type(value.dtype):
+                raise ValueError(
+                    f"{self.feature_type} is a discrete feature type therefore dtype of data array "
+                    f"has to be either integer or boolean type but feature {feature_name} has dtype {value.dtype.type}."
+                )
 
-            #         raise ValueError(f'{self.feature_type} is a discrete feature type therefore dtype of data has to '
-            #                          f'be a subtype of numpy.integer or numpy.bool, found type {value.dtype.type}')
-            # # This checking is disabled for now
-            # else:
-            #     if not issubclass(value.dtype.type, (np.floating, np.float)):
-            #         raise ValueError(f'{self.feature_type} is a floating feature type therefore dtype of data has to '
-            #                          f'be a subtype of numpy.floating or numpy.float, found type {value.dtype.type}')
             return value
 
         if self.is_vector:
@@ -803,14 +691,17 @@ class _FeatureDict(dict):
                 value = gpd.GeoDataFrame(dict(geometry=value), crs=value.crs)
 
             if isinstance(value, gpd.GeoDataFrame):
-                if self.feature_type is FeatureType.VECTOR:
-                    if FeatureType.TIMESTAMP.value.upper() not in value:
-                        raise ValueError(f"{self.feature_type} feature has to contain a column 'TIMESTAMP' with "
-                                         "timestamps")
+                if self.feature_type is FeatureType.VECTOR and FeatureType.TIMESTAMP.value.upper() not in value:
+                    raise ValueError(
+                        f"{self.feature_type} feature has to contain a column 'TIMESTAMP' with timestamps but "
+                        f"feature {feature_name} doesn't not have it."
+                    )
 
                 return value
 
-            raise ValueError(f'{self.feature_type} feature works with data of type {gpd.GeoDataFrame.__name__}, '
-                             f'parsing data type {type(value)} is not supported')
+            raise ValueError(
+                f"{self.feature_type} feature works with data of type {gpd.GeoDataFrame.__name__} but feature "
+                f"{feature_name} has data of type {type(value)}."
+            )
 
         return value
